from __future__ import annotations

from collections.abc import Awaitable
from contextlib import AsyncExitStack
from functools import partial
from inspect import isawaitable
from typing import Callable

from anyio import (
    TASK_STATUS_IGNORED,
    Event,
    create_memory_object_stream,
    create_task_group,
)
from anyio.abc import TaskGroup, TaskStatus
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from pycrdt import (
    Doc,
    YMessageType,
    YSyncMessageType,
    create_sync_message,
    create_update_message,
    handle_sync_message,
)
from structlog import BoundLogger, get_logger

from .awareness import Awareness
from .websocket import Websocket
from .ystore import BaseYStore
from .yutils import put_updates


class YRoom:
    clients: list
    ydoc: Doc
    ystore: BaseYStore | None
    _on_message: Callable[[bytes, Websocket], Awaitable[bool] | bool] | None
    _update_send_stream: MemoryObjectSendStream
    _update_receive_stream: MemoryObjectReceiveStream
    _ready: bool
    _task_group: TaskGroup | None
    _started: Event | None
    _starting: bool

    def __init__(
        self,
        ydoc: Doc | None = None,
        ready: bool = True,
        ystore: BaseYStore | None = None,
        log: BoundLogger | None = None,
    ):
        """Initialize the object.

        The YRoom instance should preferably be used as an async context manager:
        ```py
        async with room:
            ...
        ```
        However, a lower-level API can also be used:
        ```py
        task = asyncio.create_task(room.start())
        await room.started.wait()
        ...
        room.stop()
        ```

        Arguments:
            ready: Whether the internal YDoc is ready to be synchronized right away.
            ystore: An optional store in which to persist document updates.
            log: An optional logger.
        """
        self.ydoc = Doc() if ydoc is None else ydoc
        self.awareness = Awareness(self.ydoc)
        self._update_send_stream, self._update_receive_stream = create_memory_object_stream(
            max_buffer_size=65536
        )
        self._ready = False
        self.ready = ready
        self.ystore = ystore
        self.log = log or get_logger()
        self.clients = []
        self._on_message = None
        self._started = None
        self._starting = False
        self._task_group = None

    @property
    def started(self):
        """An async event that is set when the YRoom provider has started."""
        if self._started is None:
            self._started = Event()
        return self._started

    @property
    def ready(self) -> bool:
        """
        Returns:
            True is the internal YDoc is ready to be synchronized.
        """
        return self._ready

    @ready.setter
    def ready(self, value: bool) -> None:
        """
        Arguments:
            value: True if the internal YDoc is ready to be synchronized, False otherwise."""
        self._ready = value
        if value:
            self.ydoc.observe(partial(put_updates, self._update_send_stream))

    @property
    def on_message(self) -> Callable[[bytes, Websocket], Awaitable[bool] | bool] | None:
        """
        Returns:
            The optional callback to call when a message is received.
        """
        return self._on_message

    @on_message.setter
    def on_message(self, value: Callable[[bytes, Websocket], Awaitable[bool] | bool] | None):
        """
        Arguments:
            value: An optional callback to call when a message is received.
                If the callback returns True, the message is skipped.
        """
        self._on_message = value

    async def _broadcast_updates(self):
        if self.ystore is not None and not self.ystore.started.is_set():
            self._task_group.start_soon(self.ystore.start)

        async with self._update_receive_stream:
            async for update in self._update_receive_stream:
                if self._task_group.cancel_scope.cancel_called:
                    return
                # broadcast internal ydoc's update to all clients, that includes changes from the
                # clients and changes from the backend (out-of-band changes)
                for client in self.clients:
                    self.log.debug(
                        "Sending Y update to client",
                        endpoint=client.path,
                    )
                    message = create_update_message(update)
                    self._task_group.start_soon(client.send, message)
                if self.ystore:
                    self.log.debug("Writing Y update to YStore")
                    self._task_group.start_soon(self.ystore.write, update)

    async def __aenter__(self) -> YRoom:
        if self._task_group is not None:
            raise RuntimeError("YRoom already running")

        async with AsyncExitStack() as exit_stack:
            tg = create_task_group()
            self._task_group = await exit_stack.enter_async_context(tg)
            self._exit_stack = exit_stack.pop_all()
            tg.start_soon(self._broadcast_updates)
            self.started.set()

        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        if self._task_group is None:
            raise RuntimeError("YRoom not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the room.

        Arguments:
            task_status: The status to set when the task has started.
        """
        if self._starting:
            return
        else:
            self._starting = True

        if self._task_group is not None:
            raise RuntimeError("YRoom already running")

        async with create_task_group() as self._task_group:
            self._task_group.start_soon(self._broadcast_updates)
            self.started.set()
            self._starting = False
            task_status.started()

    def stop(self):
        """Stop the room."""
        if self._task_group is None:
            raise RuntimeError("YRoom not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None

    async def serve(self, websocket: Websocket):
        async with create_task_group() as tg:
            self.clients.append(websocket)
            sync_message = create_sync_message(self.ydoc)
            self.log.debug(
                "Sending message",
                name=YSyncMessageType.SYNC_STEP1.name,
                endpoint=websocket.path,
            )
            await websocket.send(sync_message)
            try:
                async for message in websocket:
                    # filter messages (e.g. awareness)
                    skip = False
                    if self.on_message:
                        _skip = self.on_message(message, websocket)
                        skip = await _skip if isawaitable(_skip) else _skip
                    if skip:
                        continue
                    message_type = message[0]
                    if message_type == YMessageType.SYNC:
                        # update our internal state in the background
                        # changes to the internal state are then forwarded to all clients
                        # and stored in the YStore (if any)
                        reply = handle_sync_message(message[1:], self.ydoc)
                        if reply is not None:
                            self.log.debug(
                                "Sending message",
                                name=YSyncMessageType.SYNC_STEP2.name,
                                endpoint=websocket.path,
                            )
                            tg.start_soon(websocket.send, reply)
                    elif message_type == YMessageType.AWARENESS:
                        # forward awareness messages from this client to all clients,
                        # including itself, because it's used to keep the connection alive
                        self.log.debug(
                            "Received message",
                            name=YMessageType.AWARENESS.name,
                            endpoint=websocket.path,
                        )
                        for client in self.clients:
                            self.log.debug(
                                "Sending Y awareness",
                                from_endpoint=websocket.path,
                                to_endpoint=client.path,
                            )
                            tg.start_soon(client.send, message)
            except Exception as e:
                self.log.debug(
                    "Error serving",
                    endpoint=websocket.path,
                    exc_info=e,
                )

            # remove this client
            self.clients = [c for c in self.clients if c != websocket]



from __future__ import annotations

from typing import TypedDict

from channels.generic.websocket import AsyncWebsocketConsumer  # type: ignore
from pycrdt import Doc, YMessageType, YSyncMessageType, create_sync_message, handle_sync_message
from structlog import get_logger

from .websocket import Websocket

logger = get_logger()


class _WebsocketShim(Websocket):
    def __init__(self, path, send_func) -> None:
        self._path = path
        self._send_func = send_func

    @property
    def path(self) -> str:
        return self._path

    def __aiter__(self):
        raise NotImplementedError()

    async def __anext__(self) -> bytes:
        raise NotImplementedError()

    async def send(self, message: bytes) -> None:
        await self._send_func(message)

    async def recv(self) -> bytes:
        raise NotImplementedError()


class YjsConsumer(AsyncWebsocketConsumer):
    """A working consumer for [Django Channels](https://github.com/django/channels).

    This consumer can be used out of the box simply by adding:
    ```py
    path("ws/<str:room>", YjsConsumer.as_asgi())
    ```
    to your `urls.py` file. In practice, once you
    [set up Channels](https://channels.readthedocs.io/en/1.x/getting-started.html),
    you might have something like:
    ```py
    # urls.py
    from django.urls import path
    from backend.consumer import DocConsumer, UpdateConsumer

    urlpatterns = [
        path("ws/<str:room>", YjsConsumer.as_asgi()),
    ]

    # asgi.py
    import os
    from channels.routing import ProtocolTypeRouter, URLRouter
    from urls import urlpatterns

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "backend.settings")

    application = ProtocolTypeRouter({
        "websocket": URLRouter(urlpatterns_ws),
    })
    ```

    Additionally, the consumer can be subclassed to customize its behavior.

    In particular,

    - Override `make_room_name` to customize the room name.
    - Override `make_ydoc` to initialize the YDoc. This is useful to initialize it with data
      from your database, or to add observers to it).
    - Override `connect` to do custom validation (like auth) on connect,
      but be sure to call `await super().connect()` in the end.
    - Call `group_send_message` to send a message to an entire group/room.
    - Call `send_message` to send a message to a single client, although this is not recommended.

    A full example of a custom consumer showcasing all of these options is:
    ```py
    from pycrdt import Doc
    from asgiref.sync import async_to_sync
    from channels.layers import get_channel_layer
    from ypy_websocket.django_channels_consumer import YjsConsumer
    from ypy_websocket.yutils import create_update_message


    class DocConsumer(YjsConsumer):
        def make_room_name(self) -> str:
            # modify the room name here
            return self.scope["url_route"]["kwargs"]["room"]

        async def make_ydoc(self) -> Doc:
            doc = Doc()
            # fill doc with data from DB here
            doc.observe(self.on_update_event)
            return doc

        async def connect(self):
            user = self.scope["user"]
            if user is None or user.is_anonymous:
                await self.close()
                return
            await super().connect()

        def on_update_event(self, event):
            # process event here
            ...

        async def doc_update(self, update_wrapper):
            update = update_wrapper["update"]
            self.ydoc.apply_update(update)
            await self.group_send_message(create_update_message(update))


    def send_doc_update(room_name, update):
        layer = get_channel_layer()
        async_to_sync(layer.group_send)(room_name, {"type": "doc_update", "update": update})
    ```

    """

    def __init__(self):
        super().__init__()
        self.room_name = None
        self.ydoc = None
        self._websocket_shim = None

    def make_room_name(self) -> str:
        """Make the room name for a new channel.

        Override to customize the room name when a channel is created.

        Returns:
            The room name for a new channel. Defaults to the room name from the URL route.
        """
        return self.scope["url_route"]["kwargs"]["room"]

    async def make_ydoc(self) -> Doc:
        """Make the YDoc for a new channel.

        Override to customize the YDoc when a channel is created
        (useful to initialize it with data from your database, or to add observers to it).

        Returns:
            The YDoc for a new channel. Defaults to a new empty YDoc.
        """
        return Doc()

    def _make_websocket_shim(self, path: str) -> _WebsocketShim:
        return _WebsocketShim(path, self.group_send_message)

    async def connect(self) -> None:
        self.room_name = self.make_room_name()
        self.ydoc = await self.make_ydoc()
        self._websocket_shim = self._make_websocket_shim(self.scope["path"])

        await self.channel_layer.group_add(self.room_name, self.channel_name)
        await self.accept()

        sync_message = create_sync_message(self.ydoc)
        logger.debug(
            "Sending message",
            name=YSyncMessageType.SYNC_STEP1.name,
            endpoint=self._websocket_shim.path,
        )
        await self._websocket_shim.send(sync_message)

    async def disconnect(self, code) -> None:
        await self.channel_layer.group_discard(self.room_name, self.channel_name)

    async def receive(self, text_data=None, bytes_data=None):
        if bytes_data is None:
            return
        await self.group_send_message(bytes_data)
        if bytes_data[0] != YMessageType.SYNC:
            return
        reply = handle_sync_message(bytes_data[1:], self.ydoc)
        if reply is not None:
            logger.debug(
                "Sending message",
                name=YSyncMessageType.SYNC_STEP2.name,
                endpoint=self._websocket_shim.path,
            )
            await self._websocket_shim.send(reply)

    class WrappedMessage(TypedDict):
        """A wrapped message to send to the client."""

        message: bytes

    async def send_message(self, message_wrapper: WrappedMessage) -> None:
        """Send a message to the client.

        Arguments:
            message_wrapper: The message to send, wrapped.
        """
        await self.send(bytes_data=message_wrapper["message"])

    async def group_send_message(self, message: bytes) -> None:
        """Send a message to the group.

        Arguments:
            message: The message to send.
        """
        await self.channel_layer.group_send(
            self.room_name, {"type": "send_message", "message": message}
        )



from typing import Protocol


class Websocket(Protocol):
    """WebSocket.

    The Websocket instance can receive messages using an async iterator,
    until the connection is closed:
    ```py
    async for message in websocket:
        ...
    ```
    Or directly by calling `recv()`:
    ```py
    message = await websocket.recv()
    ```
    Sending messages is done with `send()`:
    ```py
    await websocket.send(message)
    ```
    """

    @property
    def path(self) -> str:
        """The WebSocket path."""
        ...

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        try:
            message = await self.recv()
        except Exception:
            raise StopAsyncIteration()

        return message

    async def send(self, message: bytes) -> None:
        """Send a message.

        Arguments:
            message: The message to send.
        """
        ...

    async def recv(self) -> bytes:
        """Receive a message.

        Returns:
            The received message.
        """
        ...



import pytest
from anyio import create_task_group
from fps import get_root_module
from httpx import AsyncClient
from httpx_ws import aconnect_ws
from structlog.testing import capture_logs

from jupyverse_api.asgi_websocket_transport import ASGIWebSocketTransport
from jupyverse_api.yjs.models import CreateDocumentSession


@pytest.mark.anyio
async def test_concurrent_disconnect(tmp_path):
    config = {
        "jupyverse": {
            "type": "jupyverse",
                "config": {
                    "start_server": False,
                },
            "modules": {
                "app": {
                    "type": "app",
                },
                "yjs": {
                    "type": "yjs",
                    "config": {
                        "document_cleanup_delay": 0,
                    },
                },
                "auth": {
                    "type": "noauth",
                },
                "contents": {
                    "type": "contents",
                },
                "file_id": {
                    "type": "file_id",
                },
                "file_watcher": {
                    "type": "file_watcher",
                },
                "frontend": {
                    "type": "frontend",
                },
            },
        }
    }

    with capture_logs() as cap_logs:
        async with get_root_module(config) as root_module:
            app = root_module.app
            transport = ASGIWebSocketTransport(app=app)
            async with AsyncClient(transport=transport, base_url="http://testserver") as client:
                data = CreateDocumentSession(format="text", type="file")
                p = tmp_path / "hello.txt"
                p.write_text("hello")
                response = await client.put(
                    f"http://testserver/api/collaboration/session/{p}",
                    json=dict(data),
                )
                data = response.json()
                file_id = data["fileId"]

                async def connect_ws():
                    async with aconnect_ws(
                        f"http://testserver/api/collaboration/room/text:file:{file_id}",
                        client,
                    ):
                        pass

                async with create_task_group() as tg:
                    tg.start_soon(connect_ws)
                    tg.start_soon(connect_ws)

    assert {
        "event": "Application failed",
        "log_level": "critical",
    } not in cap_logs



from .asgi_server import ASGIServer as ASGIServer
from .websocket_provider import WebsocketProvider as WebsocketProvider
from .websocket_server import WebsocketServer as WebsocketServer
from .websocket_server import YRoom as YRoom



from __future__ import annotations

from anyio import TASK_STATUS_IGNORED, Event, create_task_group
from anyio.abc import TaskGroup, TaskStatus
from pycrdt import Doc
from structlog import BoundLogger, get_logger

from .websocket import Websocket
from .yroom import YRoom


class WebsocketServer:
    auto_clean_rooms: bool
    rooms: dict[str, YRoom]
    _task_group: TaskGroup

    def __init__(
        self,
        rooms_ready: bool = True,
        auto_clean_rooms: bool = True,
        log: BoundLogger | None = None,
    ) -> None:
        self.rooms_ready = rooms_ready
        self.auto_clean_rooms = auto_clean_rooms
        self.log = log or get_logger()
        self.rooms = {}

    async def get_room(self, name: str, ydoc: Doc | None = None) -> YRoom:
        if name not in self.rooms.keys():
            self.rooms[name] = YRoom(ydoc=ydoc, ready=self.rooms_ready, log=self.log)
        room = self.rooms[name]
        await self.start_room(room)
        return room

    async def start_room(self, room: YRoom) -> None:
        if not room.started.is_set():
            await self._task_group.start(room.start)

    def get_room_name(self, room: YRoom) -> str:
        return list(self.rooms.keys())[list(self.rooms.values()).index(room)]

    def rename_room(
        self, to_name: str, *, from_name: str | None = None, from_room: YRoom | None = None
    ) -> None:
        if from_name is not None and from_room is not None:
            raise RuntimeError("Cannot pass from_name and from_room")
        if from_name is None:
            assert from_room is not None
            from_name = self.get_room_name(from_room)
        self.rooms[to_name] = self.rooms.pop(from_name)

    def delete_room(self, *, name: str | None = None, room: YRoom | None = None) -> None:
        if name is not None and room is not None:
            raise RuntimeError("Cannot pass name and room")
        if name is None:
            assert room is not None
            name = self.get_room_name(room)
        room = self.rooms.pop(name)
        room.stop()

    async def serve(self, websocket: Websocket, stop_event: Event | None = None) -> None:
        async with create_task_group() as tg:
            tg.start_soon(self._serve, websocket, tg)
            if stop_event is not None:
                tg.start_soon(self._watch_stop, tg, stop_event)

    async def _watch_stop(self, tg: TaskGroup, stop_event: Event):
        await stop_event.wait()
        tg.cancel_scope.cancel()

    async def _serve(self, websocket: Websocket, tg: TaskGroup):
        room = await self.get_room(websocket.path)
        await self.start_room(room)
        await room.serve(websocket)

        if self.auto_clean_rooms and not room.clients:
            self.delete_room(room=room)
        tg.cancel_scope.cancel()

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        # create the task group and wait forever
        async with create_task_group() as tg:
            self._task_group = tg
            tg.start_soon(Event().wait)
            task_status.started()

    async def stop(self) -> None:
        self._task_group.cancel_scope.cancel()



from __future__ import annotations

from anyio import create_task_group
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.contents import Contents
from jupyverse_api.file_id import FileId
from jupyverse_api.main import Lifespan
from jupyverse_api.yjs import Yjs, YjsConfig

from .routes import _Yjs


class YjsModule(Module):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.config = YjsConfig(**kwargs)

    async def prepare(self) -> None:
        self.put(self.config, YjsConfig)

        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]
        contents = await self.get(Contents)  # type: ignore[type-abstract]
        file_id = await self.get(FileId)  # type: ignore[type-abstract]
        lifespan = await self.get(Lifespan)

        self.yjs = _Yjs(app, auth, contents, file_id, lifespan, self.config)

        async with create_task_group() as tg:
            await tg.start(self.yjs.start)
            self.put(self.yjs, Yjs, teardown_callback=self.yjs.stop)
            self.done()



from __future__ import annotations

from pathlib import Path

import anyio
from anyio.streams.memory import MemoryObjectSendStream
from pycrdt import TransactionEvent


def put_updates(update_send_stream: MemoryObjectSendStream, event: TransactionEvent) -> None:
    try:
        update = event.update
        update_send_stream.send_nowait(update)
    except Exception:
        pass


async def get_new_path(path: str) -> str:
    p = Path(path)
    ext = p.suffix
    p_noext = p.with_suffix("")
    i = 1
    dir_list = [p async for p in anyio.Path().iterdir()]
    while True:
        new_path = f"{p_noext}({i}){ext}"
        if new_path not in dir_list:
            break
        i += 1
    return str(new_path)



from __future__ import annotations

from datetime import datetime
from functools import partial
from uuid import uuid4

import structlog
from anyio import TASK_STATUS_IGNORED, create_task_group, sleep
from anyio.abc import TaskStatus
from anyioutils import ResourceLock, Task, create_task
from fastapi import (
    HTTPException,
    Request,
    Response,
    WebSocketDisconnect,
    status,
)
from jupyter_ydoc import ydocs as YDOCS
from jupyter_ydoc.ybasedoc import YBaseDoc
from pycrdt import Doc, YMessageType, YSyncMessageType

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.contents import Contents
from jupyverse_api.file_id import FileId
from jupyverse_api.main import Lifespan
from jupyverse_api.yjs import Yjs, YjsConfig
from jupyverse_api.yjs.models import CreateDocumentSession

from .ywebsocket.websocket import Websocket
from .ywebsocket.websocket_server import WebsocketServer, YRoom
from .ywebsocket.ystore import SQLiteYStore, YDocNotFound
from .ywidgets import Widgets

YFILE = YDOCS["file"]
AWARENESS = 1
SERVER_SESSION = uuid4().hex
logger = structlog.get_logger()


class JupyterSQLiteYStore(SQLiteYStore):
    db_path = ".jupyter_ystore.db"  # FIXME: pass in config


class _Yjs(Yjs):
    def __init__(
        self,
        app: App,
        auth: Auth,
        contents: Contents,
        file_id: FileId,
        lifespan: Lifespan,
        config: YjsConfig,
    ) -> None:
        super().__init__(app=app, auth=auth)
        self.contents = contents
        self.file_id = file_id
        self.lifespan = lifespan
        self.config = config
        if Widgets is None:
            self.widgets = None
        else:
            self.widgets = Widgets()  # type: ignore

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        async with create_task_group() as tg:
            self.room_manager = RoomManager(self.contents, self.file_id, self.lifespan, self.config)
            tg.start_soon(self.room_manager.start)
            task_status.started()

    async def stop(self) -> None:
        await self.room_manager.stop()

    async def collaboration_room_websocket(
        self,
        path,
        websocket_permissions,
    ):
        if websocket_permissions is None:
            return
        websocket, permissions = websocket_permissions
        await websocket.accept()
        ywebsocket = YWebsocket(websocket, path)
        await self.room_manager.serve(ywebsocket, permissions)

    async def create_roomid(
        self,
        path,
        request: Request,
        response: Response,
        user: User,
    ):
        # we need to process the request manually
        # see https://github.com/tiangolo/fastapi/issues/3373#issuecomment-1306003451
        create_document_session = CreateDocumentSession(**(await request.json()))
        idx = await self.file_id.get_id(path)
        res = {
            "format": create_document_session.format,
            "type": create_document_session.type,
            "sessionId": SERVER_SESSION,
        }
        if idx is not None:
            res["fileId"] = idx
            return res

        idx = await self.file_id.index(path)
        if idx is None:
            raise HTTPException(status_code=404, detail=f"File {path} does not exist")

        response.status_code = status.HTTP_201_CREATED
        res["fileId"] = idx
        return res

    def get_document(self, document_id: str) -> YBaseDoc:
        return self.room_manager.documents[document_id]


def to_datetime(iso_date: str) -> datetime:
    return datetime.fromisoformat(iso_date.rstrip("Z"))


class YWebsocket:
    """An wrapper to make a Starlette's WebSocket look like a ywebsocket's WebSocket"""

    def __init__(self, websocket, path: str):
        self._websocket = websocket
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            message = await self._websocket.receive_bytes()
        except WebSocketDisconnect:
            raise StopAsyncIteration()
        return message

    async def send(self, message):
        try:
            await self._websocket.send_bytes(message)
        except BaseException:
            return

    async def recv(self):
        return await self._websocket.receive_bytes()


class RoomManager:
    contents: Contents
    file_id: FileId
    lifespan: Lifespan
    documents: dict[str, YBaseDoc]
    watchers: dict[str, Task]
    savers: dict[str, Task]
    cleaners: dict[YRoom, Task]
    last_modified: dict[str, datetime]
    websocket_server: JupyterWebsocketServer
    room_write_permissions: dict[str, set[YWebsocket]]
    room_lock: ResourceLock

    def __init__(self, contents: Contents, file_id: FileId, lifespan: Lifespan, config: YjsConfig):
        self.contents = contents
        self.file_id = file_id
        self.lifespan = lifespan
        self.documents = {}  # a dictionary of room_name:document
        self.watchers = {}  # a dictionary of file_id:task
        self.savers = {}  # a dictionary of file_id:task
        self.cleaners = {}  # a dictionary of room:task
        self.last_modified = {}  # a dictionary of file_id:last_modification_date
        self.room_write_permissions = {}  # a dictionary of room_name:websockets that can write
        self.websocket_server = JupyterWebsocketServer(rooms_ready=False, auto_clean_rooms=False)
        self.room_lock = ResourceLock()
        self.config = config

    async def start(self):
        async with create_task_group() as self.task_group:
            await self.task_group.start(self.websocket_server.start)
            await self.lifespan.shutdown_request.wait()
            await self.websocket_server.stop()

    async def stop(self):
        for task in (
            list(self.watchers.values()) + list(self.savers.values()) + list(self.cleaners.values())
        ):
            task.cancel(raise_exception=False)

    async def serve(self, websocket: YWebsocket, permissions) -> None:
        async with self.room_lock(websocket.path):
            room = await self.websocket_server.get_room(websocket.path)
            can_write = permissions is None or "write" in permissions.get("yjs", [])
            if websocket.path not in self.room_write_permissions:
                self.room_write_permissions[websocket.path] = set()
            if can_write:
                self.room_write_permissions[websocket.path].add(websocket)

            room.on_message = self.filter_message
            is_stored_document = websocket.path.count(":") >= 2
            if is_stored_document:
                assert room.ystore is not None
                file_format, file_type, file_id = websocket.path.split(":", 2)
                if room in self.cleaners:
                    # cleaning the room was scheduled because there was no client left
                    # cancel that since there is a new client
                    self.cleaners[room].cancel(raise_exception=False)
                    await self.cleaners[room].wait()
                    if room in self.cleaners:
                        del self.cleaners[room]
                if not room.ready:
                    file_path = await self.file_id.get_path(file_id)
                    assert file_path is not None
                    logger.info(
                        "Opening collaboration room",
                        room_id=websocket.path,
                        file_path=file_path,
                    )
                    document = YDOCS.get(file_type, YFILE)(room.ydoc)
                    document.file_id = file_id
                    self.documents[websocket.path] = document
                    model = await self.contents.read_content(file_path, True, file_format)
                    assert model.last_modified is not None
                    self.last_modified[file_id] = to_datetime(model.last_modified)
                    if not room.ready:
                        # try to apply Y updates from the YStore for this document
                        try:
                            await room.ystore.apply_updates(room.ydoc)
                            read_from_source = False
                        except YDocNotFound:
                            # YDoc not found in the YStore, create the document from
                            # the source file (no change history)
                            read_from_source = True
                        if not read_from_source:
                            # if YStore updates and source file are out-of-sync, resync updates
                            # with source
                            if document.source != model.content:
                                read_from_source = True
                        if read_from_source:
                            document.source = model.content
                            await room.ystore.encode_state_as_update(room.ydoc)

                        document.dirty = False
                        room.ready = True
                        # save the document to file when changed
                        document.observe(
                            partial(
                                self.on_document_change,
                                file_id,
                                file_type,
                                file_format,
                                document,
                            )
                        )
                        # update the document when file changes
                        if file_id not in self.watchers:
                            self.watchers[file_id] = create_task(
                                self.watch_file(file_format, file_id, document),
                                self.task_group,
                            )

        await self.websocket_server.serve(websocket, self.lifespan.shutdown_request)

        # Obtain lock for cleanup to prevent race conditions when multiple clients disconnect
        async with self.room_lock(websocket.path):
            if websocket in self.room_write_permissions.get(websocket.path, set()):
                self.room_write_permissions[websocket.path].remove(websocket)

            if (
                not self.lifespan.shutdown_request.is_set()
                and is_stored_document
                and not room.clients
                and room not in self.cleaners
            ):
                # no client in this room after we disconnect
                self.cleaners[room] = create_task(
                    self.maybe_clean_room(room, websocket.path),
                    self.task_group,
                )

    async def filter_message(self, message: bytes, websocket: Websocket) -> bool:
        """
        Called whenever a message is received, before forwarding it to other clients.

        :param message: received message.
        :param websocket: the websocket that sent the message.
        :returns: True if the message must be discarded, False otherwise (default: False).
        """
        skip = False
        byte = message[0]
        msg = message[1:]
        can_write = websocket in self.room_write_permissions.get(websocket.path, set())
        if byte == AWARENESS:
            # changes = self.room.awareness.get_changes(msg)
            # # filter out message depending on changes
            # skip = True
            pass
        elif byte == YMessageType.SYNC:
            if not can_write and msg[0] == YSyncMessageType.SYNC_UPDATE:
                skip = True
        else:
            skip = True

        return skip

    async def get_file_path(self, file_id: str, document) -> str | None:
        file_path = await self.file_id.get_path(file_id)
        if file_path is None:
            return None
        if file_path != document.path:
            document.path = file_path
        return file_path

    async def watch_file(self, file_format: str, file_id: str, document: YBaseDoc) -> None:
        file_path = await self.get_file_path(file_id, document)
        assert file_path is not None
        logger.info("Watching file", path=file_path)
        # FIXME: handle file rename/move?
        watcher = self.file_id.watch(file_path)
        async for changes in watcher:
            new_file_path = await self.get_file_path(file_id, document)
            if new_file_path is None:
                continue
            if new_file_path != file_path:
                # file was renamed
                self.file_id.unwatch(file_path, watcher)
                file_path = new_file_path
                # break
            await self.maybe_load_file(file_format, file_path, file_id)
        if file_id in self.watchers:
            del self.watchers[file_id]

    async def maybe_load_file(self, file_format: str, file_path: str, file_id: str) -> None:
        model = await self.contents.read_content(file_path, False)
        # do nothing if the file was saved by us
        assert model.last_modified is not None
        if self.last_modified[file_id] < to_datetime(model.last_modified):
            # the file was not saved by us, update the shared document(s)
            model = await self.contents.read_content(file_path, True, file_format)
            assert model.last_modified is not None
            documents = [v for k, v in self.documents.items() if k.split(":", 2)[2] == file_id]
            for document in documents:
                document.source = model.content
            self.last_modified[file_id] = to_datetime(model.last_modified)

    def on_document_change(
        self, file_id: str, file_type: str, file_format: str, document: YBaseDoc, target, event
    ) -> None:
        if target == "state" and "dirty" in event.keys:
            dirty = event.keys["dirty"]["newValue"]
            if not dirty:
                # we cleared the dirty flag, nothing to save
                return
        # unobserve and observe again because the structure of the document may have changed
        # e.g. a new cell added to a notebook
        document.unobserve()
        document.observe(
            partial(self.on_document_change, file_id, file_type, file_format, document)
        )
        if file_id in self.savers:
            self.savers[file_id].cancel(raise_exception=False)
        self.savers[file_id] = create_task(
            self.maybe_save_document(file_id, file_type, file_format, document),
            self.task_group,
        )

    async def maybe_save_document(
        self, file_id: str, file_type: str, file_format: str, document: YBaseDoc
    ) -> None:
        # save after configured delay to prevent too frequent saving
        await sleep(self.config.document_save_delay)
        # if the room cannot be found, don't save
        try:
            file_path = await self.get_file_path(file_id, document)
        except Exception:
            return
        assert file_path is not None
        model = await self.contents.read_content(file_path, True, file_format)
        assert model.last_modified is not None
        if self.last_modified[file_id] < to_datetime(model.last_modified):
            # file changed on disk, let's revert
            document.source = model.content
            self.last_modified[file_id] = to_datetime(model.last_modified)
            return
        if model.content != document.source:
            # don't save if not needed
            # this also prevents the dirty flag from bouncing between windows of
            # the same document opened as different types (e.g. notebook/text editor)
            content = {
                "content": document.source,
                "format": file_format,
                "path": file_path,
                "type": file_type,
            }
            await self.contents.write_content(content)
            model = await self.contents.read_content(file_path, False)
            assert model.last_modified is not None
            self.last_modified[file_id] = to_datetime(model.last_modified)
        document.dirty = False
        # we're done saving, remove the saver
        if file_id in self.savers:
            del self.savers[file_id]

    async def maybe_clean_room(self, room, ws_path: str) -> None:
        file_id = ws_path.split(":", 2)[2]
        # keep the document for a while in case someone reconnects
        await sleep(self.config.document_cleanup_delay)

        async with self.room_lock(ws_path):
            document = self.documents[ws_path]
            document.unobserve()
            del self.documents[ws_path]
            documents = [v for k, v in self.documents.items() if k.split(":", 2)[2] == file_id]
            if not documents:
                self.watchers[file_id].cancel(raise_exception=False)
                await self.watchers[file_id].wait()
                if file_id in self.watchers:
                    del self.watchers[file_id]
            room_name = self.websocket_server.get_room_name(room)
            self.websocket_server.delete_room(room=room)

            if ws_path in self.room_write_permissions:
                del self.room_write_permissions[ws_path]

            file_path = await self.get_file_path(file_id, document)
            logger.info("Closing collaboration room", room_id=room_name, file_path=file_path)
            if room in self.cleaners:
                del self.cleaners[room]


class JupyterWebsocketServer(WebsocketServer):
    async def get_room(self, ws_path: str, ydoc: Doc | None = None) -> YRoom:
        if ws_path not in self.rooms:
            if ws_path.count(":") >= 2:
                # it is a stored document (e.g. a notebook)
                file_format, file_type, file_id = ws_path.split(":", 2)
                updates_file_path = f".{file_type}:{file_id}.y"
                ystore = JupyterSQLiteYStore(path=updates_file_path)  # FIXME: pass in config
                self.rooms[ws_path] = YRoom(ydoc=ydoc, ready=False, ystore=ystore)
            else:
                # it is a transient document (e.g. awareness)
                self.rooms[ws_path] = YRoom(ydoc=ydoc)
        room = self.rooms[ws_path]
        await self.start_room(room)
        return room



from __future__ import annotations

import struct
import tempfile
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable
from contextlib import AsyncExitStack
from inspect import isawaitable
from pathlib import Path
from typing import Callable, cast

import anyio
from anyio import TASK_STATUS_IGNORED, Event, Lock, create_task_group
from anyio.abc import TaskGroup, TaskStatus
from pycrdt import Decoder, Doc, write_var_uint
from sqlite_anyio import connect
from structlog import BoundLogger, get_logger

from .yutils import get_new_path


class YDocNotFound(Exception):
    pass


class BaseYStore(ABC):
    metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None
    version = 2
    _started: Event | None = None
    _starting: bool = False
    _task_group: TaskGroup | None = None

    @abstractmethod
    def __init__(
        self, path: str, metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None
    ): ...

    @abstractmethod
    async def write(self, data: bytes) -> None: ...

    @abstractmethod
    async def read(self) -> AsyncIterator[tuple[bytes, bytes]]: ...

    @property
    def started(self) -> Event:
        if self._started is None:
            self._started = Event()
        return self._started

    async def __aenter__(self) -> BaseYStore:
        if self._task_group is not None:
            raise RuntimeError("YStore already running")

        async with AsyncExitStack() as exit_stack:
            tg = create_task_group()
            self._task_group = await exit_stack.enter_async_context(tg)
            self._exit_stack = exit_stack.pop_all()
            await tg.start(self.start)

        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        await self.stop()
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the store.

        Arguments:
            task_status: The status to set when the task has started.
        """
        if self._starting:
            return

        self._starting = True

        if self._task_group is not None:
            raise RuntimeError("YStore already running")

        self.started.set()
        self._starting = False
        task_status.started()

    async def stop(self) -> None:
        """Stop the store."""
        if self._task_group is None:
            raise RuntimeError("YStore not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None

    async def get_metadata(self) -> bytes:
        """
        Returns:
            The metadata.
        """
        if self.metadata_callback is None:
            return b""

        metadata = self.metadata_callback()
        if isawaitable(metadata):
            metadata = await metadata
        metadata = cast(bytes, metadata)
        return metadata

    async def encode_state_as_update(self, ydoc: Doc) -> None:
        """Store a YDoc state.

        Arguments:
            ydoc: The YDoc from which to store the state.
        """
        update = ydoc.get_update()
        await self.write(update)

    async def apply_updates(self, ydoc: Doc) -> None:
        """Apply all stored updates to the YDoc.

        Arguments:
            ydoc: The YDoc on which to apply the updates.
        """
        async for update, *rest in self.read():  # type: ignore
            ydoc.apply_update(update)


class FileYStore(BaseYStore):
    """A YStore which uses one file per document."""

    path: str
    metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None
    lock: Lock

    def __init__(
        self,
        path: str,
        metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None,
        log: BoundLogger | None = None,
    ) -> None:
        """Initialize the object.

        Arguments:
            path: The file path used to store the updates.
            metadata_callback: An optional callback to call to get the metadata.
            log: An optional logger.
        """
        self.path = path
        self.metadata_callback = metadata_callback
        self.log = log or get_logger()
        self.lock = Lock()

    async def check_version(self) -> int:
        """Check the version of the store format.

        Returns:
            The offset where the data is located in the file.
        """
        if not await anyio.Path(self.path).exists():
            version_mismatch = True
        else:
            version_mismatch = False
            move_file = False
            async with await anyio.open_file(self.path, "rb") as f:
                header = await f.read(8)
                if header == b"VERSION:":
                    version = int(await f.readline())
                    if version == self.version:
                        offset = await f.tell()
                    else:
                        version_mismatch = True
                else:
                    version_mismatch = True
                if version_mismatch:
                    move_file = True
            if move_file:
                new_path = await get_new_path(self.path)
                self.log.warning(
                    "YStore version mismatch, moving file",
                    from_path=self.path,
                    to_path=new_path,
                )
                await anyio.Path(self.path).rename(new_path)
        if version_mismatch:
            async with await anyio.open_file(self.path, "wb") as f:
                version_bytes = f"VERSION:{self.version}\n".encode()
                await f.write(version_bytes)
                offset = len(version_bytes)
        return offset

    async def read(self) -> AsyncIterator[tuple[bytes, bytes, float]]:  # type: ignore
        """Async iterator for reading the store content.

        Returns:
            A tuple of (update, metadata, timestamp) for each update.
        """
        async with self.lock:
            if not await anyio.Path(self.path).exists():
                raise YDocNotFound
            offset = await self.check_version()
            async with await anyio.open_file(self.path, "rb") as f:
                await f.seek(offset)
                data = await f.read()
                if not data:
                    raise YDocNotFound
        i = 0
        for d in Decoder(data).read_messages():
            if i == 0:
                update = d
            elif i == 1:
                metadata = d
            else:
                timestamp = struct.unpack("<d", d)[0]
                yield update, metadata, timestamp
            i = (i + 1) % 3

    async def write(self, data: bytes) -> None:
        """Store an update.

        Arguments:
            data: The update to store.
        """
        parent = Path(self.path).parent
        async with self.lock:
            await anyio.Path(parent).mkdir(parents=True, exist_ok=True)
            await self.check_version()
            async with await anyio.open_file(self.path, "ab") as f:
                data_len = write_var_uint(len(data))
                await f.write(data_len + data)
                metadata = await self.get_metadata()
                metadata_len = write_var_uint(len(metadata))
                await f.write(metadata_len + metadata)
                timestamp = struct.pack("<d", time.time())
                timestamp_len = write_var_uint(len(timestamp))
                await f.write(timestamp_len + timestamp)


class TempFileYStore(FileYStore):
    """A YStore which uses the system's temporary directory.
    Files are writen under a common directory.
    To prefix the directory name (e.g. /tmp/my_prefix_b4whmm7y/):

    ```py
    class PrefixTempFileYStore(TempFileYStore):
        prefix_dir = "my_prefix_"
    ```
    """

    prefix_dir: str | None = None
    base_dir: str | None = None

    def __init__(
        self,
        path: str,
        metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None,
        log: BoundLogger | None = None,
    ):
        """Initialize the object.

        Arguments:
            path: The file path used to store the updates.
            metadata_callback: An optional callback to call to get the metadata.
            log: An optional logger.
        """
        full_path = str(Path(self.get_base_dir()) / path)
        super().__init__(full_path, metadata_callback=metadata_callback, log=log)

    def get_base_dir(self) -> str:
        """Get the base directory where the update file is written.

        Returns:
            The base directory path.
        """
        if self.base_dir is None:
            self.make_directory()
        assert self.base_dir is not None
        return self.base_dir

    def make_directory(self):
        """Create the base directory where the update file is written."""
        type(self).base_dir = tempfile.mkdtemp(prefix=self.prefix_dir)


class SQLiteYStore(BaseYStore):
    """A YStore which uses an SQLite database.
    Unlike file-based YStores, the Y updates of all documents are stored in the same database.

    Subclass to point to your database file:

    ```py
    class MySQLiteYStore(SQLiteYStore):
        db_path = "path/to/my_ystore.db"
    ```
    """

    db_path: str = "ystore.db"
    # Determines the "time to live" for all documents, i.e. how recent the
    # latest update of a document must be before purging document history.
    # Defaults to never purging document history (None).
    document_ttl: int | None = None
    path: str
    lock: Lock
    db_initialized: Event

    def __init__(
        self,
        path: str,
        metadata_callback: Callable[[], Awaitable[bytes] | bytes] | None = None,
        log: BoundLogger | None = None,
    ) -> None:
        """Initialize the object.

        Arguments:
            path: The file path used to store the updates.
            metadata_callback: An optional callback to call to get the metadata.
            log: An optional logger.
        """
        self.path = path
        self.metadata_callback = metadata_callback
        self.log = log or get_logger()
        self.lock = Lock()
        self.db_initialized = Event()

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the SQLiteYStore.

        Arguments:
            task_status: The status to set when the task has started.
        """
        self._db = await connect(self.db_path)
        await self._init_db()
        await super().start(task_status=task_status)

    async def stop(self) -> None:
        """Stop the store."""
        await self._db.close()
        await super().stop()

    async def _init_db(self):
        create_db = False
        move_db = False
        if not await anyio.Path(self.db_path).exists():
            create_db = True
        else:
            async with self.lock:
                cursor = await self._db.cursor()
                await cursor.execute(
                    "SELECT count(name) FROM sqlite_master WHERE type='table' and name='yupdates'"
                )
                table_exists = (await cursor.fetchone())[0]
                if table_exists:
                    await cursor.execute("pragma user_version")
                    version = (await cursor.fetchone())[0]
                    if version != self.version:
                        move_db = True
                        create_db = True
                else:
                    create_db = True
        if move_db:
            new_path = await get_new_path(self.db_path)
            self.log.warning(
                "YStore version mismatch, moving database",
                from_path=self.db_path,
                to_path=new_path,
            )
            await anyio.Path(self.db_path).rename(new_path)
        if create_db:
            async with self.lock:
                cursor = await self._db.cursor()
                await cursor.execute(
                    "CREATE TABLE yupdates "
                    "(path TEXT NOT NULL, yupdate BLOB, metadata BLOB, timestamp REAL NOT NULL)"
                )
                await cursor.execute(
                    "CREATE INDEX idx_yupdates_path_timestamp ON yupdates (path, timestamp)"
                )
                await cursor.execute(f"PRAGMA user_version = {self.version}")
                await self._db.commit()
        self.db_initialized.set()

    async def read(self) -> AsyncIterator[tuple[bytes, bytes, float]]:  # type: ignore
        """Async iterator for reading the store content.

        Returns:
            A tuple of (update, metadata, timestamp) for each update.
        """
        await self.db_initialized.wait()
        try:
            async with self.lock:
                cursor = await self._db.cursor()
                await cursor.execute(
                    "SELECT yupdate, metadata, timestamp FROM yupdates WHERE path = ?",
                    (self.path,),
                )
                found = False
                for update, metadata, timestamp in await cursor.fetchall():
                    found = True
                    yield update, metadata, timestamp
                if not found:
                    raise YDocNotFound
        except Exception:
            raise YDocNotFound

    async def write(self, data: bytes) -> None:
        """Store an update.

        Arguments:
            data: The update to store.
        """
        await self.db_initialized.wait()
        async with self.lock:
            # first, determine time elapsed since last update
            cursor = await self._db.cursor()
            await cursor.execute(
                "SELECT timestamp FROM yupdates WHERE path = ? ORDER BY timestamp DESC LIMIT 1",
                (self.path,),
            )
            row = await cursor.fetchone()
            diff = (time.time() - row[0]) if row else 0

            if self.document_ttl is not None and diff > self.document_ttl:
                # squash updates
                ydoc: Doc = Doc()
                await cursor.execute("SELECT yupdate FROM yupdates WHERE path = ?", (self.path,))
                for (update,) in await cursor.fetchall():
                    ydoc.apply_update(update)
                # delete history
                await cursor.execute("DELETE FROM yupdates WHERE path = ?", (self.path,))
                # insert squashed updates
                squashed_update = ydoc.get_update()
                metadata = await self.get_metadata()
                await cursor.execute(
                    "INSERT INTO yupdates VALUES (?, ?, ?, ?)",
                    (self.path, squashed_update, metadata, time.time()),
                )

            # finally, write this update to the DB
            metadata = await self.get_metadata()
            await cursor.execute(
                "INSERT INTO yupdates VALUES (?, ?, ?, ?)",
                (self.path, data, metadata, time.time()),
            )
            await self._db.commit()
